// Linetrim.cpp : Defines the entry point for the console application.
//

// LDraw line trimmer
// Trim lines that meet triangle or quads. Trimmed part is the one where triangle 
// or quad is seen CCW
// Usage: linetrim LdrawLineFile LdrawTriQuadFile LdrawTrimmedLineFileOut [LdrawLineRemovedFile]
// Optional LdrawLineRemovedFile is a copy of input LdrawLineFile with all lines removed.
// Only Trimming occurs only with triangles or quads,
// primitives must be inlined before usage if needed.

// Philippe "Philo" Hurbain - 2007 - www.philohome.com

// Contains excerpts of triangle intersection code from Tomas Mller
// See thesis here: http://www.cs.lth.se/home/Tomas_Akenine_Moller/pubs/tritri.pdf
// Complete code here: http://jgt.akpeters.com/papers/Moller97/tritri.html


// Contains softSurfer Line - Triangle intersection (adaptation)
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

#include "stdafx.h"

/* Triangle/triangle intersection test routine,
 * by Tomas Moller, 1997.
 * See article "A Fast Triangle-Triangle Intersection Test",
 * Journal of Graphics Tools, 2(2), 1997
 * updated: 2001-06-20 (added line of intersection)
 *
 * This version computes the line of intersection as well (if they are not coplanar):
 * int tri_tri_intersect_with_isectline(double V0[3],double V1[3],double V2[3], 
 *				        double U0[3],double U1[3],double U2[3],int *coplanar,
 *				        double isectpt1[3],double isectpt2[3]);
 * coplanar returns whether the tris are coplanar
 * isectpt1, isectpt2 are the endpoints of the line of intersection
 */

#include <math.h>


#define USE_EPSILON_TEST TRUE  
#define EPSILON 0.000001

/* Null vector */

double null[3] = {0,0,0};

/* some macros */

#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2]; 

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2]; 

#define MANHATTAN(v1, v2) (fabs(v1[0]-v2[0]) + fabs(v1[1]-v2[1]) + fabs(v1[2]-v2[2]))


int tri_segment_intersect_with_isectpoint(double V0[3],double V1[3],double V2[3],
										  double S0[3],double S1[3],
										  double isectpt[3], int *idx)
{
  double u[3],v[3],n[3]; // triangle vectors
  double dir[3], w0[3], w[3];// ray vectors
  double temp[3];
  double r, a, b;
  double d1;
  double du0,du1;
  double du0du1;
  
  /* compute plane equation of triangle(V0,V1,V2) */
  SUB(u,V1,V0);
  SUB(v,V2,V0);
  CROSS(n,u,v);
  d1=-DOT(n,V0);
  /* plane equation 1: n.X+d1=0 */

  /* put S0, S1 into plane equation 1 to compute signed distances to the plane*/
  du0=DOT(n,S0)+d1;
  du1=DOT(n,S1)+d1;
  
  /* coplanarity robustness check */
#if USE_EPSILON_TEST==TRUE
  if(fabs(du0)<EPSILON) du0=0.0;
  if(fabs(du1)<EPSILON) du1=0.0;
#endif
  du0du1=du0*du1;

  if(du0du1>0.0f) /* same sign on both of them + not equal 0 ? */
    return 0;                    /* no intersection occurs */

  if(MANHATTAN(n,null)<EPSILON) // triangle is degenerate
	return 0;

  *idx=(du0>0)?0:1; // Which segment end is "outside" triangle?

  SUB(dir, S1, S0);  // ray direction vector
  SUB(w0, S0, V0);
  a=-DOT(n,w0);
  b=DOT(n,dir);

  if (fabs(b) < EPSILON) {     // ray is parallel to triangle plane
     return 0;             
  }

  r = a / b;
  MULT(temp, dir, r);
  ADD(isectpt, S0, temp);

  // is I inside T?
    double uu, uv, vv, wu, wv, D1;
    uu = DOT(u,u);
    uv = DOT(u,v);
    vv = DOT(v,v);
    SUB(w, isectpt, V0); 
    wu = DOT(w,u);
    wv = DOT(w,v);
    D1 = uv * uv - uu * vv;

    // get and test parametric coords
    double s, t;
    s = (uv * wv - vv * wu) / D1;
    if (s < -EPSILON || s > 1.0+EPSILON)        // I is outside T
        return 0;
    t = (uv * wu - uu * wv) / D1;
    if (t < -EPSILON || (s + t) > 1.0+EPSILON)  // I is outside T
        return 0;

    return 1;                      // I is in T
}

#define MAX_TRI 1000
#define MAX_LINE 2000

int main(int argc, char* argv[])
{
	double InTri[MAX_TRI+1][3][3];
	double InLine[MAX_LINE][2][3];

	FILE *LdrawTriQuadFile;
	FILE *LdrawLineFile;
	FILE *LdrawTrimmedLineFileOut;
	FILE *LdrawLineRemovedFile;

	int NumTri, NumLine, TrimLine;

	char buf[1024];
	int type, color;

	int i, j, k;
	
	int index;

	int flag=0;

	double isectpt[3];

	printf ("Linetrim V1.1 - by Philo \n\n");

	if((argc < 4) || (argc > 5) ) {
		printf ("Usage: Linetrim LdrawTriQuadFile LdrawLineFile LdrawTrimmedLineFileOut [LdrawLineRemovedFile]\n");
		printf("Press <Enter> to quit");
		getchar();
		return 1;
	}
	

	LdrawLineFile = fopen (argv[2], "rt");
	if(LdrawLineFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[2]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	LdrawTriQuadFile = fopen (argv[1], "rt");
	if(LdrawTriQuadFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[1]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	LdrawTrimmedLineFileOut = fopen (argv[3], "wt");
	if(LdrawTrimmedLineFileOut == NULL)
	{
		printf("Can't create file\n%s\n", argv[3]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	if (argc == 5)
	{
		LdrawLineRemovedFile = fopen (argv[4], "wt");
		if(LdrawLineRemovedFile == NULL)
		{
			printf("Can't create file\n%s\n", argv[4]);
			printf("Press <Enter> to quit");
			getchar();
			return 2;
		}
	}

//    fprintf(LdrawTrimmedLineFileOut, "0 BFC CERTIFY CW\n"); //DEBUG

	printf("\nRead Tri/Quad file...\n");
	NumTri=0;
	while (fgets(buf, 1024, LdrawTriQuadFile))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==3)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InTri[NumTri][0][0], &InTri[NumTri][0][1], &InTri[NumTri][0][2], 
				&InTri[NumTri][1][0], &InTri[NumTri][1][1], &InTri[NumTri][1][2],
				&InTri[NumTri][2][0], &InTri[NumTri][2][1], &InTri[NumTri][2][2]);
			NumTri++;
//			fprintf(LdrawTrimmedLineFileOut, "%s",buf);	// DEBUG
		}
		if(type==4)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InTri[NumTri][0][0], &InTri[NumTri][0][1], &InTri[NumTri][0][2], 
				&InTri[NumTri][1][0], &InTri[NumTri][1][1], &InTri[NumTri][1][2],
				&InTri[NumTri][2][0], &InTri[NumTri][2][1], &InTri[NumTri][2][2],
				&InTri[NumTri+1][2][0], &InTri[NumTri+1][2][1], &InTri[NumTri+1][2][2]);
//			fprintf(LdrawTrimmedLineFileOut, "3 24 %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",InTri[NumTri][0][0], InTri[NumTri][0][1], InTri[NumTri][0][2], InTri[NumTri][1][0], InTri[NumTri][1][1], InTri[NumTri][1][2], InTri[NumTri][2][0], InTri[NumTri][2][1], InTri[NumTri][2][2]);	// DEBUG


			SET(InTri[NumTri+1][0], InTri[NumTri][0]);
			SET(InTri[NumTri+1][1], InTri[NumTri][2]);

//			fprintf(LdrawTrimmedLineFileOut, "3 24 %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",InTri[NumTri+1][0][0], InTri[NumTri+1][0][1], InTri[NumTri+1][0][2], InTri[NumTri+1][1][0], InTri[NumTri+1][1][1], InTri[NumTri+1][1][2], InTri[NumTri+1][2][0], InTri[NumTri+1][2][1], InTri[NumTri+1][2][2]);	// DEBUG

			NumTri+=2;

		}

		if(NumTri>MAX_TRI)
		{
			printf("Too many triangles in file 1\n");
			printf("Press <Enter> to quit");
			getchar();
			return 3;
		}

	}
	printf("%d triangles in Tri/Quad file\n", NumTri);

	printf("\nRead Line file...\n");
	NumLine=0;
	while (fgets(buf, 1024, LdrawLineFile))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==2)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InLine[NumLine][0][0], &InLine[NumLine][0][1], &InLine[NumLine][0][2], 
				&InLine[NumLine][1][0], &InLine[NumLine][1][1], &InLine[NumLine][1][2]);
			NumLine++;
		}
		else if(argc == 5)
		{
			fprintf(LdrawLineRemovedFile, "%s", buf);
		}
		if(NumLine>MAX_TRI)
		{
			printf("Too many lines in file 2\n");
			printf("Press <Enter> to quit");
			getchar();
			return 3;
		}

	}
	printf("%d lines in file 2\n", NumLine);

	printf("Computing intersections\n");
	TrimLine=0;

	for(i=0; i<NumLine; i++)
	{
		for(j=0; j<NumTri; j++)
		{

			if(tri_segment_intersect_with_isectpoint(InTri[j][0], InTri[j][1], InTri[j][2],
				InLine[i][0], InLine[i][1], isectpt, &index))
			{
				SET(InLine[i][index], isectpt);
				TrimLine++;
			}
		}
		printf(".");
	}
	printf("\n%d trim(s) performed\n", TrimLine);

	printf("Writing output file\n");

	fprintf(LdrawTrimmedLineFileOut, "0 Lines %s trimmed by %s\n", argv[2], argv[1]);
	fprintf(LdrawTrimmedLineFileOut, "0 Name: %s\n", argv[3]);
	fprintf(LdrawTrimmedLineFileOut, "0 Author: Philo's Line Trimmer\n0\n");

	for(k=0; k<NumLine; k++)
	{
		fprintf(LdrawTrimmedLineFileOut, "2 24 %f %f %f %f %f %f\n", InLine[k][0][0], InLine[k][0][1], InLine[k][0][2], InLine[k][1][0], InLine[k][1][1], InLine[k][1][2]);
	}
	printf("%d line(s) written\n", NumLine);
	fprintf(LdrawTrimmedLineFileOut, "0\n");
	fclose (LdrawTriQuadFile);
	fclose (LdrawLineFile);
	fclose (LdrawTrimmedLineFileOut);
	if (argc == 5) fclose (LdrawLineRemovedFile);
	return(0);
}
